Statistical_maps = {
    'General settings': {'type': 'subheading'},

    'verbose': {'type': 'CheckButton', 'Recommended': 'true', 'label': 'Verbose statistical map stages',
                'Description': 'true or false. Print progress to terminal.'},

    'multicore_processing': {'type': 'CheckButton', 'Recommended': 'true',
                             'Description': 'true or false. Use multicore processing during analysis? '
                                            'Multicore processing currently works within participants not between them.'
                                            ' Recommended: true'},

    'max_core_usage': {'type': 'OptionMenu', 'Recommended': 'max', 'Options': ['max', 6, 5, 4, 3, 2, 1],
                       'save_as': 'string',
                       'Description': "'max' to select number of cores available on the system, alternatively an int to"
                                      " manually select number of cores to use. Recommended: 'max'"},

    'base_folder': {'type': 'Entry', 'Recommended': "", 'save_as': 'string',
                    'label': 'Base folder location',
                    'Description': 'Either the absolute location of the folder containing the subjects or blank, '
                                   'if blank then a browser window will allow you to search for the files at '
                                   'runtime.'},

    'input_folder_name': {'type': 'Entry', 'Recommended': "func", 'save_as': 'string',
                          'Description': 'Folder found in each subjects directory containing the files to be analysed.'
                          },

    'output_folder_name': {'type': 'Entry', 'Recommended': 'DEFAULT', 'save_as': 'string',
                           'Description': 'Directory to save output. If set to DEFAULT, the default name for the '
                                          'statistical map created will be used. '
                                          'Recommended: DEFAULT'},

    'High pass filtering': {'type': 'subheading'},

    'temporal_filter': {'type': 'CheckButton', 'Recommended': 'true',
                        'Description': 'true or false. Use a high pass filter to remove low frequency drift. '
                                       'Recommended: true'},

    'highpass_filter_cutoff': {'type': 'Entry', 'Recommended': 0.01,
                               'label': 'Highpass filter cutoff frequency (Hz)',
                               'Description': 'Highpass filter cutoff frequency converted into sigma in seconds using '
                                              'the formula 1/(2*f*TR). Recommended: 0.01'},

    'Motion correction': {'type': 'subheading'},

    'remove_motion_outliers': {'type': 'CheckButton', 'Recommended': 'true',
                               'Description': 'true or false. Use fsl_motion_outliers to remove motion outliers'
                                              '(uses default fsl_motion_outliers settings). Recommended: true'},

    'motion_correction': {'type': 'CheckButton', 'Recommended': 'true',
                          'Description': 'true or false. Use MCFLIRT to motion correct volumes '
                                         '(uses default MCFLIRT settings). Recommended: true'},

    'Spatial smoothing': {'type': 'subheading'},

    'spatial_smoothing': {'type': 'CheckButton', 'Recommended': 'false',
                          'Description': 'true or false. Uses SUSAN to spatial smooth. Recommended: true'},

    'smoothing_fwhm': {'type': 'Entry', 'Recommended': 8.0,
                       'label': 'Spatial smoothing fwhm (mm)',
                       'Description': 'fwhm of smoothing, in mm, gets converted using sqrt(8*log(2)). Recommended: 8.0'},

    'smoothing_brightness_threshold': {'type': 'Entry', 'Recommended': 2000.0,
                                       'label': 'Spatial smoothing brightness threshold',
                                       'Description': 'Should be greater than noise level and less than contrast of '
                                                      'edges to be preserved. Recommended: 2000.0'},

    'Image SNR calculation': {'type': 'subheading'},

    'magnitude_correction': {'type': 'CheckButton', 'Recommended': 'true',
                             'Description': 'true or false. Correction factor of 0.7 applied when running iSNR '
                                            'calculations, to correct for Rayleigh distributed noise when using '
                                            'magnitude vs complex images. \nReference: Constantinides, C. D., Atalar, '
                                            'E., & McVeigh, E. R. (1997). Signal-to-Noise Measurements in Magnitude '
                                            'Images from NMR Phased Arrays.'},

    'noise_volume': {'type': 'CheckButton', 'Recommended': 'false', 'label': 'Noise volume included in time series',
                     'Description': 'true or false. Select true if a noise volume has been collected as part of the fMRI '
                                    'time series.\n'
                                    'NOTE: If true, the noise volume in the time series will be separated from the '
                                    'functional volumes and will be placed into the folder "func_noiseVolumeRemoved".\n'
                                    'If "noise volume" is true and "noise value" is not none, the noise volume '
                                    'will be used in image SNR calculation rather than the user defined noise value.'},

    'noise_volume_location': {'type': 'OptionMenu', 'Recommended': 'End', 'Options': ['Beginning', 'End'],
                              'save_as': 'string',
                              'Description': "'max' to select number of cores available on the system, alternatively "
                                             "an int to manually select number of cores to use. Recommended: 'max'"},

    'iSNR_std_use_only_nonzero_voxels': {'type': 'CheckButton', 'Recommended': 'true',
                                         'label': 'Use only nonzero voxels for iSNR calc',
                                         'Description': 'true or false.'},

    'Add Gaussian noise': {'type': 'subheading'},

    'create_noise_level_file': {'type': 'Button', 'Command': 'create_noise_file', 'Text': 'Create noiseValues.csv',
                                'Pass self': False,
                                'Description': 'Create a noiseValues.csv file to determine what the standard deviation'
                                               "of the Gaussian should be when adding noise to each participant's data "
                                               '(the distribution will have a mean of 0).'
                                               'This will not overwrite the original files. '
                                               'Recommended: The standard deviation over time found in '
                                               'the dataset.'
                                               '\nNOTE: Can be used to see how additional noise affects '
                                               'analysis. To calculate the noise level for each participant, '
                                               'create tSNR maps with the fRAT and run the '
                                               'full analysis using the tStd files. The mean values of '
                                               'this analysis in each participants subfolder, will then show you the '
                                               'average noise for each region for each participant.'
                                               '\n\nThis file can also be used to manually set a noise value for each '
                                               'participant for use in iSNR calculation. This noise value can be '
                                               'calculated as the standard deviation of voxel values outside of the '
                                               'brain. If "Noise volume included in time series" is set to true, '
                                               'standard deviation of noise will be calculated using the noise volume, '
                                               'even if a noise value has been provided in this file.'},

    'noise_multipliers': {'type': 'Entry', 'Recommended': 1, 'save_as': 'list', 'label': 'Noise multiplier(s)',
                          'Description': "Provide a comma-separated list of multipliers for the standard deviation of "
                                         "the gaussian noise to plot e.g. '1, 5'. A separate file will be produced for each multiplier."
                                         "\nNOTE: Noise has a gaussian distribution, with a mean of 0 and a "
                                         "standard deviation of the noise level of each participant * multiplier."}
}



import platform
import toml
import tkinter as tk
import tkinter.ttk as ttk

import ast
import sys
import textwrap

from PIL import ImageTk
from operator import itemgetter
from tkinter import font

from fRAT.nogui import fRAT
from fRAT._version import __version__

from fRAT.utils import *
from fRAT.utils.directory_comparison import *
from fRAT.utils.fRAT_config_setup import *
from fRAT.utils.statmap_config_setup import *

from fRAT.utils import dash_report
from fRAT.utils.printResults import printResults
from fRAT.utils.statmap import main as statmap_calc

from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE, SIG_DFL)

w = None
WIDGET_Y_PADDING = 9
WIDGET_X_PADDING = 10

def start_gui():
    '''Starting point when module is the main routine.'''
    global val, w, root, top

    path_check()

    Utils.checkversion(__version__)

    print('----------------------------\n----- Running fRAT_GUI -----\n----------------------------')

    root = tk.Tk()
    top = GUI(root)

    root.mainloop()


class GUI:
    def __init__(self, window=None, page='Home', load_initial_values=True):
        '''This class configures and populates the toplevel window.
           top is the toplevel containing window.'''
        self.widgets = {}
        self.dynamic_widgets = {}
        self.frames = []
        self.page = page
        self.background = '#d9d9d9'

        if load_initial_values:
            # Load config file last used
            with open(f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/latest_settings.toml',
                      'r') as tomlfile:
                parse = tomlfile.readlines()
                parse = toml.loads(''.join(parse))

            for toml_file in [f'{folder}/{config_file}' for folder, config_file in parse.items()]:
                self.load_initial_values(toml_file)

            self.style = ttk.Style()
            self.style.theme_use('clam')

            window.geometry("+100+100")
            window.resizable(True, True)
            window.title("fRAT GUI")
            window.configure(background=self.background)
            window.configure(highlightbackground=self.background)
            window.configure(highlightcolor="black")

            self.default_font = None
            self.important_font = None
            self.subheading_font = None
            self.heading_font = None

            self.setup_fonts()

        self.change_page_specific_settings(window)

        if page == 'Statistical_maps':
            self.current_info = Statistical_maps
        else:
            try:
                self.current_info = eval(self.page)
            except NameError:
                self.current_info = pages

        self.create_settings_frame(window)

        if self.page == 'Home':
            self.banner_draw(window)
            self.create_home_options_frame(self.settings_frame)
            self.create_home_atlas_frame(self.settings_frame)
            self.create_homestar_runner_frame(self.settings_frame)
            self.create_statmap_frame(window)

    def setup_fonts(self):
        self.default_font = font.Font(font='TkDefaultFont')
        self.important_font = font.Font(font='TkDefaultFont')
        self.heading_font = font.Font(font='TkDefaultFont')
        self.subheading_font = font.Font(font='TkDefaultFont')

        self.important_font.configure(weight='bold')

        if platform.system() == 'Darwin':
            self.heading_font.configure(weight='bold', size=20)
            self.subheading_font.configure(weight='bold', size=18)
        else:
            self.heading_font.configure(weight='bold', size=15)
            self.subheading_font.configure(weight='bold', size=12)

    def change_page_specific_settings(self, window):
        statistics_width = "530"
        home_width = "512"
        parsing_width = "450"

        if self.page == 'Home':
            window.geometry(f"{home_width}x860")
        elif self.page == 'General':
            window.geometry(f"{home_width}x860")
        elif self.page == 'Analysis':
            window.geometry(f"{home_width}x860")
        elif self.page == 'Parsing':
            window.geometry(f"{parsing_width}x280")
        elif self.page == 'Plotting':
            window.geometry(f"{home_width}x860")
        elif self.page == 'Statistics':
            window.geometry(f"{statistics_width}x860")
        elif self.page == 'Statistical_maps':
            window.geometry(f"{home_width}x860")

    def create_statmap_frame(self, window):
        self.statmap_frame = tk.LabelFrame(window)
        self.frames.append(self.statmap_frame)

        self.statmap_frame.place(relx=0.02, rely=0.77, relheight=0.22, relwidth=0.973)
        self.statmap_frame.configure(text='Statistical maps', font=self.heading_font)
        self.format_frame(self.statmap_frame, borderwidth=1)

        self.statmap_run_frame_create(self.statmap_frame)
        self.statmap_settings_frame_create(self.statmap_frame)

    def statmap_settings_frame_create(self, window):
        self.statmap_settings_frame = tk.LabelFrame(window)
        self.frames.append(self.statmap_settings_frame)
        self.statmap_settings_frame.place(relx=0.02, rely=0, height=140, relwidth=0.467)
        self.statmap_settings_frame.configure(text='Settings', font=self.heading_font)
        self.format_frame(self.statmap_settings_frame, borderwidth=1)

        self.statmap_save_button = ttk.Button(self.statmap_settings_frame)
        self.statmap_save_button.place(relx=0.02, rely=0.08, height=42, width=105)
        self.statmap_save_button.configure(
            command=lambda: Save_settings(['Statistical_maps'], 'maps/statmap_config.toml'), text='''Save settings''')
        Tooltip.CreateToolTip(self.statmap_save_button, 'Save all statistical map settings')

        self.statmap_reset_button = ttk.Button(self.statmap_settings_frame)
        self.statmap_reset_button.place(relx=0.5, rely=0.08, height=42, width=105)
        self.statmap_reset_button.configure(command=lambda: Reset_settings(['Statistical_maps']),
                                            text='''Reset settings''')
        Tooltip.CreateToolTip(self.statmap_reset_button, 'Reset statistical map settings')

        self.statmap_settings_button = ttk.Button(self.statmap_settings_frame)
        self.statmap_settings_button.place(relx=0.17, rely=0.55, height=42, width=150)
        self.statmap_settings_button.configure(command=lambda: self.change_frame('Statistical_maps'),
                                               text='''Settings''')

    def statmap_run_frame_create(self, window):
        self.statmap_run_frame = tk.LabelFrame(window)
        self.frames.append(self.statmap_run_frame)
        self.statmap_run_frame.place(relx=0.5, rely=0, height=140, relwidth=0.467)
        self.statmap_run_frame.configure(text='Run', font=self.heading_font)
        self.format_frame(self.statmap_run_frame, borderwidth=1)

        statmap_options = ('Image SNR', 'Temporal SNR', 'Add Gaussian noise')
        state = tk.StringVar()
        state.set(statmap_options[1])
        self.statmap_option = tk.OptionMenu(self.statmap_run_frame, state, *statmap_options)
        self.statmap_option.place(relx=0.08, rely=0.27, width=200, bordermode='ignore')
        self.statmap_option.configure(bg=self.background)
        self.statmap_option.val = state

        self.statmap_run_button = ttk.Button(self.statmap_run_frame)
        self.statmap_run_button.place(relx=0.17, rely=0.55, height=42, width=150)
        self.statmap_run_button.configure(command=lambda: button_handler('Make maps', self.statmap_option.val.get()))
        self.statmap_run_button.configure(text='''Make maps''')
        Tooltip.CreateToolTip(self.statmap_run_button, 'Make statistical maps using method chosen in the drop down '
                                                       'menu.')

    @staticmethod
    def load_initial_values(toml_file):
        with open(f"{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/{toml_file}", 'r') as f:
            for line in f.readlines():
                if line[:2] == '##':  # Subheadings
                    continue

                elif line[0] == '#':  # Headings
                    curr_page = re.split('# |\n', line)[1]

                elif curr_page == 'Version Info':
                    continue

                elif line != '\n':
                    setting = [x.replace("'", "").strip() for x in re.split(" = |\[|\]|\n|(?<!')#.*", line) if x]

                    current_setting = Utils.convert_toml_input_to_python_object(setting[1])

                    try:
                        eval(curr_page)[setting[0]]['Current'] = current_setting
                    except KeyError:
                        warnings.warn(
                            f'"{setting[0]}" not present in config setup file. Configuration file may be outdated.')

    def banner_draw(self, window):
        img = Image.open(f'{Path(os.path.abspath(__file__)).parents[0]}/images/fRAT.gif')

        zoom = 0.8
        pixels_x, pixels_y = tuple([int(zoom * x) for x in img.size])
        img = img.resize((pixels_x, pixels_y))

        width, height = img.size
        new_width = width + 88
        result = Image.new('RGB', (new_width, height - 20), (0, 0, 0))
        result.paste(img, (44, -10))

        result = ImageTk.PhotoImage(result)

        panel = tk.Label(window, image=result, borderwidth=0)
        panel.photo = result
        panel.place(relx=0.02, y=10)

        self.frames.append(panel)

    def create_home_options_frame(self, window):
        self.Options_frame = tk.LabelFrame(window)
        self.Options_frame.place(relx=0.5, rely=0.01, height=150, relwidth=0.345)
        self.Options_frame.configure(text=f'''Options''', font=self.heading_font)
        self.format_frame(self.Options_frame, borderwidth=1)
        self.frames.append(self.Options_frame)

        self.Save_button = ttk.Button(self.Options_frame)
        self.Save_button.place(relx=0.05, y=10, height=42, width=150)
        self.Save_button.configure(command=lambda: Save_settings(pages, 'roi_analysis/fRAT_config.toml'))
        self.Save_button.configure(text='''Save settings''')
        Tooltip.CreateToolTip(self.Save_button, 'Save all fRAT settings')

        self.Reset_button = ttk.Button(self.Options_frame)
        self.Reset_button.place(relx=0.05, y=70, height=42, width=150)
        self.Reset_button.configure(command=lambda: Reset_settings(pages))
        self.Reset_button.configure(text='''Reset settings''')
        Tooltip.CreateToolTip(self.Reset_button, 'Reset fRAT settings to recommended values')

    def create_home_atlas_frame(self, window):
        self.Atlas_frame = tk.LabelFrame(window)
        self.Atlas_frame.place(relx=0.5, rely=0.348, height=128, relwidth=0.467)
        self.Atlas_frame.configure(text=f'''Atlas information''', font=self.heading_font)
        self.format_frame(self.Atlas_frame, borderwidth=1)
        self.frames.append(self.Atlas_frame)

        atlas_options = ('Cerebellum_MNIflirt',
                         'Cerebellum_MNIfnirt',
                         'HarvardOxford-Cortical',
                         'HarvardOxford-Subcortical',
                         'JHU-labels',
                         'JHU-tracts',
                         'Juelich',
                         'MNI',
                         'SMATT',
                         'STN',
                         'Striatum-Structural',
                         'Talairach',
                         'Thalamus')

        state = tk.StringVar()
        state.set(atlas_options[2])
        self.Atlas_option = tk.OptionMenu(self.Atlas_frame, state, *atlas_options)
        self.Atlas_option.place(relx=0.08, rely=0.2, width=200, bordermode='ignore')
        self.Atlas_option.configure(bg=self.background)
        self.Atlas_option.val = state

        self.Atlas_button = ttk.Button(self.Atlas_frame)
        self.Atlas_button.place(relx=0.17, rely=0.45, height=42, width=150)
        self.Atlas_button.configure(command=lambda: Print_atlas_ROIs(self.Atlas_option.val.get()))
        self.Atlas_button.configure(text='''Print Atlas ROIs''')
        Tooltip.CreateToolTip(self.Atlas_button, 'Print ROIs from selected atlas. This can be used to find which '
                                                 'numbers to input in the "Plotting" menu to plot specific regions.'
                                                 '\nNOTE: This does not change the atlas to be used for analysis.')

    def create_homestar_runner_frame(self, window):
        self.Run_frame = tk.LabelFrame(window)
        self.Run_frame.place(relx=0.025, rely=0.69, height=145, relwidth=0.945)
        self.Run_frame.configure(text=f'''Run''', font=self.heading_font)
        self.format_frame(self.Run_frame, borderwidth=1)
        self.frames.append(self.Run_frame)

        self.paramValues_button = ttk.Button(self.Run_frame)
        self.paramValues_button.place(x=2, y=12, height=42, width=150)
        self.paramValues_button.configure(command=lambda: button_handler('Make paramValues.csv'))
        self.paramValues_button.configure(text='''Setup parameters''')
        Tooltip.CreateToolTip(self.paramValues_button,
                              'Creates a csv table with prefilled parameter info for each file. Can be set using a command line flag instead')

        self.Print_button = ttk.Button(self.Run_frame)
        self.Print_button.place(x=156, y=12, height=42, width=150)
        self.Print_button.configure(command=lambda: button_handler('Print_results'))
        self.Print_button.configure(text='''Print results''')
        Tooltip.CreateToolTip(self.Print_button, 'Print results of fRAT to the terminal')

        self.Dash_button = ttk.Button(self.Run_frame)
        self.Dash_button.place(x=309, y=12, height=42, width=150)
        self.Dash_button.configure(command=lambda: button_handler('Run_dash'))
        self.Dash_button.configure(text='''Interactive table''')
        Tooltip.CreateToolTip(self.Dash_button, 'Create an interactive table to display fRAT results')

        self.Run_button = ttk.Button(self.Run_frame)
        self.Run_button.place(x=137, y=66, height=42, width=181)
        self.Run_button.configure(command=lambda: button_handler('Run_fRAT'))
        self.Run_button.configure(text='''Run fRAT''')
        Tooltip.CreateToolTip(self.Run_button, 'Run fRAT with current settings')

    def create_settings_frame(self, window):
        if self.page == 'Home':
            self.settings_frame = tk.LabelFrame(window)
            self.frames.append(self.settings_frame)

            self.settings_frame.place(relx=0.02, rely=0.15, relheight=0.61, relwidth=0.973)
            self.settings_frame.configure(text='fRAT', font=self.heading_font)
            self.format_frame(self.settings_frame, borderwidth=1)

            self.General_settings_frame = tk.LabelFrame(self.settings_frame)
            self.frames.append(self.General_settings_frame)
            self.General_settings_frame.place(relx=0.025, rely=0.01, height=330, relwidth=0.41)
            self.General_settings_frame.configure(text=f'''Settings''', font=self.heading_font)
            self.format_frame(self.General_settings_frame, borderwidth=1)
            current_frame = self.General_settings_frame

            y_loc = 10
            relx = 0.03
            for page in pages:
                if page == 'Home':
                    continue

                if page == 'Violin_plot':
                    break

                self.page_switch_button_setup(page, current_frame, y_loc, relx)

                y_loc += 60

        else:
            self.create_scrollable_frame(window)
            self.create_widgets(previous_frame='Home')

    def create_plot_settings_frame(self, row):
        self.plot_settings_frame = tk.LabelFrame(self.settings_frame)
        self.frames.append(self.plot_settings_frame)
        self.plot_settings_frame.grid(row=row, column=0, columnspan=1)
        self.plot_settings_frame.configure(text=f'''Specific plot settings''', font=self.heading_font)
        self.format_frame(self.plot_settings_frame, borderwidth=1)

        return row + 1

    def create_scrollable_frame(self, window):
        self.sbf = ScrollbarFrame(window)
        self.sbf.pack(side="top", fill="both", expand=True)
        self.settings_frame = self.sbf.scrolled_frame

        self.frames.extend([self.sbf, self.settings_frame])

    def create_widgets(self, previous_frame, create_return_button=True):
        row = 0

        for setting in self.current_info:
            self.label_create(setting, row, self.current_info[setting])

            if self.current_info[setting]['type'] == 'Scale':
                widget = self.scale_create(setting, self.current_info[setting], row)
                self.widgets = {**self.widgets, **widget}

            elif self.current_info[setting]['type'] == 'CheckButton':
                widget = self.checkbutton_create(setting, self.current_info[setting], row)
                self.widgets = {**self.widgets, **widget}

            elif self.current_info[setting]['type'] == 'OptionMenu':
                widget = self.optionmenu_create(setting, self.current_info[setting], row)
                self.widgets = {**self.widgets, **widget}

            elif self.current_info[setting]['type'] == 'Entry':
                widget = self.entry_create(setting, self.current_info[setting], row)
                self.widgets = {**self.widgets, **widget}

            elif self.current_info[setting]['type'] == 'Button':
                widget = self.button_create(setting, self.current_info[setting], row)
                self.widgets = {**self.widgets, **widget}

            elif self.current_info[setting]['type'] == 'Dynamic':
                widget, row = self.dynamic_widget(setting, self.current_info[setting], row)
                self.dynamic_widgets = {**self.dynamic_widgets, **widget}

            row += 1

        self.settings_frame.configure(text=f'''{self.page.replace('_', ' ')}''', font=self.heading_font)
        self.format_frame(self.settings_frame)

        if create_return_button == True:
            self.return_button_create_and_format(previous_frame, row)

    def button_create(self, name, info, row):
        func = eval(info['Command'])

        if info['Pass self']:
            self.__setattr__(name, ttk.Button(self.settings_frame, command=lambda: func(self)))
        else:
            self.__setattr__(name, ttk.Button(self.settings_frame, command=lambda: func()))

        widget = getattr(self, name)
        widget.configure(text=info['Text'])

        widget.grid(row=row, column=1, pady=WIDGET_Y_PADDING, padx=WIDGET_X_PADDING, sticky='W')

        return {name: widget}

    def return_button_create_and_format(self, previous_frame, row):
        self.return_button = ttk.Button(self.settings_frame, command=lambda: self.change_frame(previous_frame))
        self.return_button.grid(column=0, row=row, columnspan=2, pady=15)
        self.return_button.configure(text=f'''Back to {previous_frame}''')

    def format_frame(self, frame, borderwidth=0):
        frame.configure(borderwidth=borderwidth)
        frame.configure(relief='raised')
        frame.configure(background=self.background)
        frame.configure(highlightbackground=self.background)
        frame.configure(highlightcolor="black")

    def page_switch_button_setup(self, page, frame, y_loc, relx):
        self.__setattr__(page, ttk.Button(frame, command=lambda: self.change_frame(page)))
        page_switch_button = getattr(self, page)
        page_switch_button.place(relx=relx, y=y_loc, width=185, height=42)
        page_switch_button.configure(text=f'''{page.replace('_', ' ')}''')

    def dynamic_widget(self, name, info, row):
        try:
            text = eval(info['Options'])['Current']
            text = [value.strip() for value in text.split(',')]
        except KeyError:
            text = ""

        dynamic_widgets = {}

        if info['subtype'] == 'OptionMenu':
            info['DynamOptions'] = [*text]

            # If not currently set by user, set to default value
            if info['Current'] not in info['DynamOptions']:
                try:
                    info['Current'] = info['DynamOptions'][info['DefaultNumber']]
                except IndexError:
                    info['Current'] = ''

            if len(info['DynamOptions']) == 1:
                info['DynamOptions'].append('')

            widget = self.optionmenu_create(name, info, row)

            dynamic_widgets = {**dynamic_widgets, **widget}

        elif info['subtype'] == 'OptionMenu2':
            if info['Current'] == 'FILL IV TYPE AS BETWEEN-SUBJECTS':
                info['Current'] = info['Options2'][info['DefaultNumber']]

            try:
                info['Current'] = info['Current'].split(', ')
            except AttributeError:
                # Error occurs if string has already been converted to list
                pass

            if len(info['Current']) != len(text):
                info['Current'] = info['Current'] * len(text)

            for counter, value in enumerate(text):
                widget = self.optionmenu_create(f"{name}_{value}", info, row, counter)

                dynamic_widgets = {**dynamic_widgets, **widget}

                # Replace current labels and create one for each row
                self.label_create(f"{info['label']} - {value}", row, info, fixed_name=True)

                row += 1

        elif info['subtype'] == 'Checkbutton':
            if info['Current'] == 'INCLUDE ALL VARIABLES':
                info['Current'] = Parsing['parameter_dict1']['Current'].split(', ')

            for value in text:
                widget = self.checkbutton_create(f"{name}_{value}", info, row)
                [widget.configure(text=value) for widget in widget.values()]

                dynamic_widgets = {**dynamic_widgets, **widget}

                row += 1

        return dynamic_widgets, row

    def label_create(self, name, row, info=None, fixed_name=False):
        self.__setattr__(name, tk.Label(self.settings_frame))
        label_name = getattr(self, name)

        if not fixed_name:
            try:
                name = info['label']
            except KeyError:
                if info['type'] != 'subheading':
                    name = name.capitalize()

        label_name.configure(background=self.background)
        label_name.configure(foreground="#000000")

        if info['type'] != 'subheading':
            Tooltip.CreateToolTip(label_name, info['Description'])
            label_name.grid(row=row, column=0, pady=WIDGET_Y_PADDING, padx=WIDGET_X_PADDING, sticky='W')
            label_name.configure(text=f'''{name.replace("_", " ")}:''', font=self.default_font)
        else:
            label_name.grid(row=row, column=0, pady=(30, 5), ipadx=WIDGET_X_PADDING, sticky='SW')
            label_name.configure(text=name, font=self.subheading_font)

        try:
            if info['status'] == 'important':
                label_name.configure(font=self.important_font)
        except KeyError:
            pass

    def scale_create(self, name, info, row):
        self.__setattr__(name, tk.Scale(self.settings_frame, from_=info['From'], to=info['To']))
        widget = getattr(self, name)

        widget.grid(row=row, column=1, pady=WIDGET_Y_PADDING, padx=WIDGET_X_PADDING, sticky='W')
        widget.configure(resolution=info['Resolution'])

        self.scale_default_settings(widget)

        widget.set(info['Current'])

        return {name: widget}

    def scale_default_settings(self, widget):
        widget.configure(activebackground="#ececec")
        widget.configure(background=self.background)
        widget.configure(bigincrement="0.05")
        widget.configure(foreground="#000000")
        widget.configure(highlightbackground=self.background)
        widget.configure(highlightcolor="black")
        widget.configure(orient="horizontal")
        widget.configure(troughcolor=self.background)

    def checkbutton_create(self, name, info, row):
        state = tk.BooleanVar()

        self.__setattr__(name, tk.Checkbutton(self.settings_frame, variable=state))
        widget = getattr(self, name)

        widget.grid(row=row, column=1, pady=WIDGET_Y_PADDING, padx=WIDGET_X_PADDING, sticky='W')
        widget.val = state

        self.checkbutton_default_settings(widget)

        if info['Current'] in ['true', 'false']:
            info['Current'] = ast.literal_eval(info['Current'].title())

        current_val = info['Current']

        # Dynamic widget handler
        if not isinstance(info['Current'], bool) and info['Current'] == 'all':
            current_val = True
        elif not isinstance(info['Current'], bool) and name.rsplit('_', 1)[1] in info['Current']:
            current_val = True
        elif not isinstance(info['Current'], bool) and not name.rsplit('_', 1)[1] in info['Current']:
            current_val = False

        if current_val:
            widget.select()
        else:
            widget.deselect()

        return {name: widget}

    def checkbutton_default_settings(self, widget):
        widget.configure(activebackground="#ececec")
        widget.configure(activeforeground="#000000")
        widget.configure(background=self.background)
        widget.configure(foreground="#000000")
        widget.configure(highlightbackground=self.background)
        widget.configure(highlightcolor="black")
        widget.configure(justify='left')

    def optionmenu_create(self, name, info, row, counter=None):
        state = tk.StringVar()
        if counter is not None:
            state.set(info['Current'][counter])
        else:
            state.set(info['Current'])

        try:
            options = info['DynamOptions']
        except KeyError:
            try:
                options = info['Options2']
            except KeyError:
                options = info['Options']

        self.__setattr__(name, tk.OptionMenu(self.settings_frame, state, *options))
        widget = getattr(self, name)

        widget.grid(row=row, column=1, pady=WIDGET_Y_PADDING, padx=WIDGET_X_PADDING, sticky='W')
        widget.configure(bg=self.background)
        widget.val = state

        return {name: widget}

    def entry_create(self, name, info, row):
        self.__setattr__(name, tk.Entry(self.settings_frame))
        widget = getattr(self, name)
        widget.grid(row=row, column=1, pady=WIDGET_Y_PADDING, padx=WIDGET_X_PADDING, sticky='W')

        self.entry_default_settings(widget)

        if info['Current'] is None:
            info['Current'] = str(info['Current'])

        if isinstance(info['Current'], (list, tuple)):
            info['Current'] = [str(x) for x in info['Current']]
            info['Current'] = ", ".join(info['Current'])

        widget.insert(0, info['Current'])

        return {name: widget}

    def entry_default_settings(self, widget):
        widget.configure(background="white")
        widget.configure(font=self.default_font)
        widget.configure(foreground="#000000")
        widget.configure(insertbackground="black")
        widget.configure(selectforeground="white")

    def change_frame(self, page):
        current_options2_menu = {}

        for widget in self.widgets:
            if eval(self.page)[widget]['type'] != 'Button':
                try:
                    eval(self.page)[widget]['Current'] = self.widgets[widget].get()
                except AttributeError:
                    eval(self.page)[widget]['Current'] = self.widgets[widget].val.get()

        for widget in self.dynamic_widgets:
            if self.dynamic_widgets[widget].winfo_class() == 'Checkbutton':
                params = widget.rsplit('_', 1)

                if isinstance(eval(self.page)[params[0]]['Current'], str):
                    eval(self.page)[params[0]]['Current'] = [x.strip() for x in
                                                             eval(self.page)[params[0]]['Current'].split(',')]

                if self.dynamic_widgets[widget].val.get():  # If checkbutton is true
                    if params[1] not in eval(self.page)[params[0]]['Current']:
                        if eval(self.page)[params[0]]['Current'][0] == '':
                            eval(self.page)[params[0]]['Current'][0] = (params[1])

                        else:
                            eval(self.page)[params[0]]['Current'].append(params[1])

                else:  # If checkbutton is false
                    if params[1] in eval(self.page)[params[0]]['Current']:
                        eval(self.page)[params[0]]['Current'].remove(params[1])

                    if len(eval(self.page)[params[0]]['Current']) == 0:
                        eval(self.page)[params[0]]['Current'].append(
                            '')  # If list is empty, set first element to blank string

            else:
                try:
                    eval(self.page)[widget]['Current'] = self.dynamic_widgets[widget].val.get()
                except KeyError:
                    # Will raise KeyError for OptionsMenu2
                    current_widget = eval(self.page)[widget.rpartition('_')[0]]

                    if current_widget['label'] in current_options2_menu:
                        current_options2_menu[current_widget['label']].append(self.dynamic_widgets[widget].val.get())
                    else:
                        current_options2_menu[current_widget['label']] = []

                    current_widget['Current'][len(current_options2_menu[current_widget['label']])] = \
                    self.dynamic_widgets[widget].val.get()

        for frame in self.frames:
            frame.destroy()

        self.frames.clear()

        self.__init__(root, page, load_initial_values=False)


class ConfigurationFiles:
    """Class container for processing configuration files."""

    analysis_config = 'fRAT_config.toml'
    statmap_config = 'statmap_config.toml'

    def update_latest_settings_file(self):
        """Update latest_settings.toml when configuration file changes"""
        pass


class Tooltip:
    def __init__(self, widget):
        self.widget = widget
        self.widget_type = widget.winfo_class()
        self.tooltip_window = None
        self.id = None
        self.x = self.y = 0

        self.font = font.Font(font='TkTooltipFont')

        if platform.system() == 'Darwin':
            self.font.configure(size=12)
        else:
            self.font.configure(size=10)

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text

        if self.tooltip_window or not self.text:
            return

        if self.widget_type == 'TButton':
            x_offset = 120
            y_offset = 32
        else:
            x_offset = 57
            y_offset = 27

        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + x_offset
        y = y + cy + self.widget.winfo_rooty() + y_offset

        self.tooltip_window = tooltip_window = tk.Toplevel(self.widget)
        tooltip_window.overrideredirect(True)
        tooltip_window.wm_geometry(f"+{x}+{y}")

        label = tk.Label(tooltip_window, text=self.text, justify=tk.LEFT,
                         background="#ffffe0", relief=tk.SOLID, borderwidth=1,
                         font=self.font)
        label.pack(ipadx=1)

        tooltip_window.update_idletasks()
        tooltip_window.lift()

    def hidetip(self):
        tooltip_window = self.tooltip_window
        self.tooltip_window = None

        if tooltip_window:
            tooltip_window.destroy()

    @staticmethod
    def CreateToolTip(widget, text):
        toolTip = Tooltip(widget)
        text = '\n'.join(['\n'.join(textwrap.wrap(line, 120, break_long_words=False, replace_whitespace=False))
                          for line in text.splitlines() if line.strip() != ''])

        def enter(event):
            toolTip.showtip(text)

        def leave(event):
            toolTip.hidetip()

        widget.bind('<Enter>', enter)
        widget.bind('<Leave>', leave)


class AutoHidingScrollbar(tk.Scrollbar):
    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            # grid_remove is currently missing from Tkinter
            self.pack_forget()
        else:
            if self.cget("orient") == 'horizontal':
                self.pack(fill='x')
            else:
                self.pack(fill='y')
        tk.Scrollbar.set(self, lo, hi)


class ScrollbarFrame(tk.LabelFrame):
    """
    Extends class tk.Frame to support a scrollable Frame
    This class is independent from the widgets to be scrolled and
    can be used to replace a standard tk.Frame
    """

    def __init__(self, parent, **kwargs):
        tk.LabelFrame.__init__(self, parent, **kwargs)

        # The Scrollbar, layout to the right
        vsb = AutoHidingScrollbar()
        vsb.pack(side="right", fill="y")

        # The Canvas which supports the Scrollbar Interface, layout to the left
        self.canvas = tk.Canvas(self, borderwidth=0, background='#d9d9d9', highlightbackground='#d9d9d9')
        self.canvas.pack(side="left", fill="both", expand=True)

        # Bind the Scrollbar to the self.canvas Scrollbar Interface
        self.canvas.configure(yscrollcommand=vsb.set)
        vsb.configure(command=self.canvas.yview)

        # The Frame to be scrolled, layout into the canvas
        # All widgets to be scrolled have to use this Frame as parent
        self.scrolled_frame = tk.LabelFrame(self.canvas, background=self.canvas.cget('bg'))
        self.canvas.create_window((0, 0), window=self.scrolled_frame, anchor='center')

        # Configures the scroll region of the Canvas dynamically
        self.scrolled_frame.bind("<Configure>", self.on_configure)

        top.frames.append(vsb)

    def on_configure(self, event):
        """Set the scroll region to encompass the scrolled frame"""
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))


def button_handler(command, *args):
    try:
        if command == 'Run_fRAT':
            check_stale_state()

            Save_settings(pages, f'roi_analysis/{ConfigurationFiles.analysis_config}')

            print('----- Running fRAT -----')
            fRAT(ConfigurationFiles.analysis_config)

        elif command == 'Print_results':
            printResults(ConfigurationFiles.analysis_config)

        elif command == "Make paramValues.csv":
            Save_settings(pages, f'roi_analysis/{ConfigurationFiles.analysis_config}')
            make_table()

        elif command == "Run_dash":
            dash_report.main(f'roi_analysis/{ConfigurationFiles.analysis_config}')

        elif command == "Make maps":
            Save_settings(['Statistical_maps'], f'maps/{ConfigurationFiles.statmap_config}')
            statmap_calc(args[0], __version__, ConfigurationFiles.statmap_config)

    except Exception as err:
        if err.args[0] == 'No folder selected.':
            print('----- Exiting -----\n')
        else:
            log = logging.getLogger(__name__)
            log.exception(err)
            sys.exit()


def run_tests(GUI):
    print('----- Running tests -----')
    # Hack to save all settings on general settings page, to be changed in the future when callbacks are implemented
    GUI.change_frame('General')
    Save_settings(pages, f'roi_analysis/{ConfigurationFiles.analysis_config}')

    path_to_example_data = find_example_dataset()

    # Create tSNR maps and run ROI analysis
    statmap_calc('Temporal SNR', __version__, 'test_config.toml', path_to_example_data)
    fRAT('test_config.toml', path_to_example_data)

    # Run tests to check if output of fRAT matches the example data
    roi_output_test = TestDifferences([f'{path_to_example_data}/sub-02/statmaps/test_maps',
                                       f'{path_to_example_data}/sub-02/statmaps/temporalSNR_report'],
                                      General['verbose_errors']['Current'])

    voxelwise_map_test = TestDifferences([f'{path_to_example_data}/test_ROI_report',
                                          f'{path_to_example_data}/HarvardOxford-Cortical_ROI_report'],
                                         General['verbose_errors']['Current'])

    # Delete files
    if General['delete_test_folder']['Current'] == 'Always' \
            or (General['delete_test_folder']['Current'] == 'If completed without error'
                and voxelwise_map_test.status == 'No errors'
                and roi_output_test.status == 'No errors'):
        shutil.rmtree(f'{path_to_example_data}/test_ROI_report')
        shutil.rmtree(f'{path_to_example_data}/sub-01/statmaps/test_maps')
        shutil.rmtree(f'{path_to_example_data}/sub-02/statmaps/test_maps')

        print('\nDeleted test folders.')

    else:
        print('\nRetaining test folders.')

    if voxelwise_map_test.status == 'No errors' and roi_output_test.status == 'No errors':
        print("\n--- End of installation testing, no errors found ---")
    else:
        warnings.warn("\n--- End of installation testing, errors found ---")


def find_example_dataset():
    example_data_folders = glob(os.path.expanduser('~/Documents/fRAT/*example_data*'))

    if not os.path.exists(os.path.expanduser('~/Documents/fRAT/')):
        raise FileNotFoundError(f'fRAT folder not present in documents directory.\n'
                                'Run mkdir ~/Documents/fRAT in the terminal, then download the example dataset from '
                                'https://osf.io/pbm3d/, extract it and place it into this folder.')

    if not example_data_folders:
        raise FileNotFoundError(f'No "example_data" folder in fRAT directory.\n'
                                f'Download it from https://osf.io/pbm3d/ and place it into the fRAT directory '
                                f'("~/Documents/fRAT/").')

    if len(example_data_folders) > 1:
        raise FileExistsError('More than one example dataset folder found in fRAT directory. '
                              'Only keep the version that matches your fRAT version.')

    return example_data_folders[0]


def check_stale_state():
    current_critical_params = Parsing['parameter_dict1']['Current'].split(', ')

    dynamic_widgets = itemgetter('table_cols', 'table_rows',
                                 'brain_table_cols', 'brain_table_rows',
                                 'single_roi_fig_colour', 'single_roi_fig_x_axis',
                                 'histogram_fig_x_facet', 'histogram_fig_y_facet',
                                 'brain_table_col_labels', 'brain_table_row_labels')(Plotting)

    dynamic_widgets += itemgetter('IV_type', 'include_as_variable')(Statistics)

    for counter, widget in enumerate(dynamic_widgets):
        if widget['Current'] == '' and len(current_critical_params) > 1:
            # If current value is blank but shouldn't be as number of critical params is above 1
            dynamic_widgets[counter]['Current'] = current_critical_params[dynamic_widgets[counter]['DefaultNumber']]

        elif widget['Recommended'] == 'CHANGE TO DESIRED LABEL' and widget['Current'] != 'CHANGE TO DESIRED LABEL':
            # If brain table column and row labels have been changed, do not change again
            # This allows these labels to be updated automatically if user never enters the brain table menu
            # But does not change to the default value if labels have been set either manually or automatically
            pass

        elif widget['Current'] == 'FILL IV TYPE AS BETWEEN-SUBJECTS':
            # Set all IV's to between subjects if not been modified already
            dynamic_widgets[counter]['Current'] = [widget['Options2'][widget['DefaultNumber']]] * len(
                current_critical_params)

        elif widget['Current'] == 'INCLUDE ALL VARIABLES':
            dynamic_widgets[counter]['Current'] = current_critical_params

        elif not widget['Current'] == '' and widget['Current'] not in current_critical_params \
                and not widget['Recommended'] in ['FILL IV TYPE AS BETWEEN-SUBJECTS', 'INCLUDE ALL VARIABLES']:
            try:
                dynamic_widgets[counter]['Current'] = current_critical_params[dynamic_widgets[counter]['DefaultNumber']]
            except IndexError:
                # If only one parameter has been given in critical params
                dynamic_widgets[counter]['Current'] = ''


def Print_atlas_ROIs(selection):
    """Extract labels from specified FSL atlas XML file."""
    try:
        fsl_path = os.environ['FSLDIR']
    except OSError:
        raise Exception('FSL environment variable not set.')

    atlas_path = f"{fsl_path}/data/atlases/{selection}.xml"

    with open(atlas_path) as fd:
        atlas_label_dict = xmltodict.parse(fd.read())

    roiArray = []
    for roiLabelLine in atlas_label_dict['atlas']['data']['label']:
        roiArray.append(roiLabelLine['#text'])

    roiArray.extend(['No ROI', 'Overall'])
    roiArray = sorted(roiArray)

    print(f"----------------------------\n{selection} Atlas:\n----------------------------")

    for roi_num, roi in enumerate(roiArray):
        print("{roi_num}: {roi}".format(roi_num=roi_num, roi=roi))

    print("----------------------------\n")


def Save_settings(page_list, file):
    with open(f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/{file}', 'w') as f:
        for page in page_list:
            if page == 'Home':
                continue

            f.write(f"# {page}\n")

            for key in eval(page).keys():
                if eval(page)[key]['type'] == 'subheading':
                    f.write(f"\n## {key}\n")
                    continue
                elif eval(page)[key]['type'] == 'Button':
                    continue

                description = eval(page)[key]['Description'].replace('\n', ' ')

                if eval(page)[key]['type'] == 'OptionMenu':
                    description += f" Options: {eval(page)[key]['Options']}."

                convert = 'default'

                try:
                    if eval(page)[key]['save_as'] == 'string':
                        convert = 'string'

                    elif eval(page)[key]['save_as'] == 'list':
                        if eval(page)[key]['Current'] == 'Runtime':
                            convert = 'string'

                        else:
                            try:
                                eval(page)[key]['Current'] = [int(x) for x in eval(page)[key]['Current'].split(',')]
                            except AttributeError:  # Handle exception if already a list so cannot split
                                pass
                            except ValueError:  # Handle exception if list items cannot be converted into ints
                                convert = 'split_list'

                    elif eval(page)[key]['save_as'] == 'string_or_list':
                        try:
                            eval(page)[key]['Current'] = list(
                                ast.literal_eval(eval(page)[key]['Current']))  # Try to convert to list

                        except (ValueError, TypeError):  # Handles error if input is None or is string
                            convert = 'string'

                except KeyError:  # Handle exception if save_as does not exist as key
                    eval(page)[key]['Current'] = str(eval(page)[key]['Current']).lower() \
                        if type(eval(page)[key]['Current']) is bool else eval(page)[key]['Current']

                    if eval(page)[key]['Current'] in ['None', None]:
                        convert = 'string'

                if convert == 'default':  # Save as is
                    offset = ' ' * (80 - len(f"{key} = {eval(page)[key]['Current']}"))
                    f.write(f"{key} = {eval(page)[key]['Current']}  "
                            f"{offset}# {description}\n")

                elif convert == 'string':  # Convert to string
                    offset = ' ' * (80 - len(f"{key} = '{eval(page)[key]['Current']}'"))
                    f.write(f"{key} = '{eval(page)[key]['Current']}'  "
                            f"{offset}# {description}\n")

                elif convert == 'split_list':  # Split items then convert to list
                    # Take out the string 'all' for Dynamic Checkbuttons, which will be the case if Recommended is set
                    # to 'all'
                    if eval(page)[key]['type'] == 'Dynamic' \
                            and eval(page)[key]['subtype'] == 'Checkbutton' \
                            and 'all,' in eval(page)[key]['Current']:
                        eval(page)[key]['Current'] = eval(page)[key]['Current'].replace('all, ', '')

                    offset = ' ' * (
                            80 - len(f"{key} = {[val.strip() for val in eval(page)[key]['Current'].split(',')]}"))
                    f.write(f"{key} = {[val.strip() for val in eval(page)[key]['Current'].split(',')]}  "
                            f"{offset}# {description}\n")

            f.write("\n")

        f.flush()
        f.close()

    print(f'----- Saved config file: {file.split("/")[-1]} -----')


def Reset_settings(pages):
    for page in pages:
        if page == 'Home':
            continue

        for key in eval(page).keys():
            if eval(page)[key]['type'] not in ['subheading', 'Button']:
                eval(page)[key]['Current'] = eval(page)[key]['Recommended']

    print('----- Reset fRAT settings to recommended values, save them to retain these settings -----')


def make_table():
    config = Utils.load_config(f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/roi_analysis',
                               ConfigurationFiles.analysis_config)  # Load config file

    print('--- Creating paramValues.csv ---')
    print('Select the base directory.')
    base_directory = Utils.file_browser(title='Select the base directory', chdir=False)

    participant_dirs = find_participant_dirs(base_directory, config)

    data = []
    for participant_dir in participant_dirs:
        if config.make_folder_structure:
            file_loc = f'{participant_dir}'

        else:
            file_loc = f'{participant_dir}/{config.parsing_folder}'

        brain_file_list = Utils.find_files(file_loc, "hdr", "nii.gz", "nii")

        if config.make_folder_structure and not brain_file_list:
            raise FileNotFoundError(f'No files found in {participant_dir}.'
                                    f'\nAs make folder structure is set to true, '
                                    f'files to be scanned should be placed into this root directory and not a '
                                    f'subdirectory.')

        brain_file_list = [Utils.strip_ext(brain) for brain in brain_file_list]
        brain_file_list.sort()

        for file in brain_file_list:
            # Try to find parameters to prefill table
            keys, brain_file_params = parse_params_from_file_name(file, config)
            data.append([os.path.split(participant_dir)[-1], file, *brain_file_params, np.NaN, np.NaN])

        if config.make_folder_structure:
            create_folder_structure(participant_dir, config)

    df = pd.DataFrame(columns=['Participant', 'File name',
                               *keys,
                               'Ignore file during analysis? (y for yes, otherwise blank)',
                               'Baseline parameter combination for statistics (y for yes, otherwise blank)'],
                      data=data)

    df.to_csv(f'{base_directory}/paramValues.csv', index=False)

    print(f"\nparamValues.csv saved in {base_directory}."
          f"\n\nMake sure values in paramValues.csv are correct before continuing analysis. Also make sure anatomy "
          f"scans were not present when setting up parameters, or these lines will need to be removed from "
          f"paramValues.csv."
          f"\nIf the csv file contains unexpected parameters, update the parsing options in the GUI and rerun setup "
          f"parameters, or manually update them.")

    if config.make_folder_structure:
        print(f"\nSet up folder structure and moved fMRI volumes into {config.parsing_folder} directory.")

    if config.automatically_create_statistics_options_file and config.parameter_dict1 != ['']:
        create_statistics_file(directory=base_directory)

    elif config.automatically_create_statistics_options_file and config.parameter_dict1 == ['']:
        print(f"\nstatisticsOptions.csv not saved, no independent variables given in parsing menu.")


def create_statistics_file(directory=''):
    Save_settings(pages, f'roi_analysis/{ConfigurationFiles.analysis_config}')

    config = Utils.load_config(f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/roi_analysis',
                               ConfigurationFiles.analysis_config)

    if not directory:
        directory = Utils.file_browser('Select base directory or report output directory')

    table, folder_type = Utils.load_paramValues_file(directory=directory)

    if folder_type == 'report_folder':
        # Go back to base folder if report folder is selected
        directory += '/..'

    data = []

    for parameter in config.parameter_dict1:
        data.append([parameter, 'Calculate main effect'])

        table.columns = [x.lower() for x in table.columns]

        unique_vals = table[parameter.lower()].unique()
        unique_vals.sort()

        combinations = list(itertools.combinations(sorted(unique_vals), 2))

        for combination in combinations:
            data.append([' v '.join(str(v) for v in combination), np.NaN])

        data.extend([[np.NaN, np.NaN],
                     [parameter, 'Calculate simple effect', 'Exclude from analysis']])

        for value in unique_vals:
            data.append([value, np.NaN, np.NaN])

        data.append([np.NaN, np.NaN, np.NaN])

    df = pd.DataFrame(data=data)
    df.to_csv(f'{directory}/statisticsOptions.csv', index=False, header=False)

    print('--- Created statisticsOptions.csv ---')


def create_noise_file():
    print('--- Creating noiseValues.csv ---')
    Utils.load_config(f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/maps',
                      ConfigurationFiles.statmap_config)

    directory = Utils.file_browser('Select base directory')
    _, participant_names = Utils.find_participant_dirs(directory=directory)

    data = []
    for participant in participant_names:
        data.append([participant, np.NaN, np.NaN])

    df = pd.DataFrame(columns=['Participant', 'Noise over time', 'Background noise'],
                      data=data)

    df.to_csv(f'{directory}/noiseValues.csv', index=False)
    print('--- Created noiseValues.csv ---')


def create_folder_structure(participant, config):
    direcs = ['func', 'anat', 'fslfast', 'statmaps']

    if config.parsing_folder not in direcs:
        direcs.append(config.parsing_folder)

    for direc in direcs:
        Utils.check_and_make_dir(f"{participant}/{direc}")

    brain_file_list = Utils.find_files(participant, "hdr", "nii.gz", "nii", "json")
    for file in brain_file_list:
        Utils.move_file(file, participant, f'{participant}/{config.parsing_folder}', rename_copy=False)


def find_participant_dirs(base_directory, config):
    participant_dirs = [direc for direc in glob(f"{base_directory}/*") if re.search("sub-[0-9]+", direc)]

    if len(participant_dirs) == 0:
        raise FileNotFoundError('Participant directories not found.\n'
                                'Make sure participant directories are labelled e.g. sub-01 and the selected '
                                'directory contains all participant directories.')
    elif config.verbose:
        print(f'Found {len(participant_dirs)} participant folders.')

    return participant_dirs


def parse_params_from_file_name(json_file_name, cfg=config):
    """Search for MRI parameters in each json file name for use in table headers and created the combined json."""
    param_nums = []
    keys = []

    for key in cfg.parameter_dict:
        parameter = cfg.parameter_dict[key]  # Extract search term

        if parameter == '':
            continue

        elif key in cfg.binary_params:
            param = re.search("{}".format(parameter), json_file_name, flags=re.IGNORECASE)

            if param is not None:
                param_nums.append('On')  # Save 'on' if parameter is found in file name
            else:
                param_nums.append('Off')  # Save 'off' if parameter not found in file name

        else:
            # Float search
            param = re.search("{}[0-9][p|.][0-9]".format(parameter), json_file_name, flags=re.IGNORECASE)
            if param is not None:
                param_nums.append(param[0][1] + "." + param[0][-1])

            # If float search didn't work then integer search
            else:
                param = re.search("{}[0-9]".format(parameter), json_file_name, flags=re.IGNORECASE)
                if param is not None:
                    param_nums.append(param[0][-1])  # Extract the number from the parameter

                else:
                    param_nums.append(str(param))  # Save None if parameter not found in file name

        keys.append(key)

    return keys, param_nums


def path_check():
    try:
        os.environ['FSLDIR']
    except KeyError:
        raise IsADirectoryError("$FSLDIR not in path")

    if os.path.isdir(f"{os.environ['FSLDIR']}/bin"):
        print(f"\nFSL directory found in {os.environ['FSLDIR']}.")
    else:
        raise IsADirectoryError(f"bin folder not found in {os.environ['FSLDIR']}")


if __name__ == '__main__':
    start_gui()



import copy
import itertools
import os
import re
import warnings
from glob import glob
from pathlib import Path

import matplotlib.image as mpimg
import numpy as np
import nibabel as nib
import pandas as pd
import plotnine as pltn
import simplejson as json
from nilearn import plotting
from matplotlib import pyplot as plt
from PIL import Image

from .utils import Utils

import matplotlib
matplotlib.use('agg')  # Use non-GUI backend to prevent memory leak caused by creating subplots in a loop


class Figures:
    config = None

    @classmethod
    def make_figures(cls, cfg, config_path, config_filename):
        cls.config = cfg

        combined_results_df, _ = Utils.read_combined_results(os.getcwd(), cls.config.averaging_type)

        if not os.path.exists('Figures'):
            os.makedirs('Figures')

        Utils.save_config('Figures', config_path, config_filename,
                          new_config_name='figure_log',
                          relevant_sections=['Parsing',
                                             'Plotting'],
                          additional_info=[f'data_used_for_figures = '
                                           f'"{cfg.averaging_type.replace(" ", "_").lower()}"'])

        if cls.config.verbose:
            print(f'\nUsing {cfg.averaging_type.replace(" ", "_").lower()} data to create figures.')

        if cls.config.multicore_processing & (cls.config.make_one_region_fig or cls.config.make_histogram):
            pool = Utils.start_processing_pool()
        else:
            pool = None

        if cls.config.make_brain_table:
            if cls.config.verbose:
                print(f'\n--- Brain grid creation ---')

            BrainGrid.make(combined_results_df)

        if cls.config.make_violin_plot:
            if cls.config.verbose:
                print(f'\n--- Violin plot creation ---')

            Utils.check_and_make_dir("Figures/Violin_plots")
            ViolinPlot.make(combined_results_df)

        if cls.config.make_one_region_fig:
            Barchart.setup(combined_results_df, pool)

        if cls.config.make_histogram:
            Histogram.setup(combined_results_df, pool)

        if pool:
            pool.close()
            pool.join()

    @staticmethod
    def find_axis_limit(thisroi, figure, axis):
        # Find what the current axis limits will be
        fig = figure.draw()
        axes = fig.axes

        if axis == 'xaxis':
            lim = axes[0].get_xlim()

        elif axis == 'yaxis':
            lim = axes[0].get_ylim()

        lim = (*lim, thisroi)

        return lim

    @staticmethod
    def figure_save(figure, thisroi, folder, chart_type, config):
        # Format file name correctly
        replacements = [
            (r'\([^()]*\)', ""),  # Remove anything between parenthesis
            (r'[^a-zA-Z\d:]', "_"),  # Remove non-alphanumeric characters
            (r'_{2,}', "_")  # Replace multiple underscores with one
        ]
        thisroi = thisroi.replace("\'", "")  # Remove apostrophes
        for old, new in replacements:
            thisroi = re.sub(old, new, thisroi)

        Utils.check_and_make_dir(f"Figures/{chart_type.title()}s/{folder}/")
        figure.save(f"Figures/{chart_type.title()}s/{folder}/{thisroi}_{chart_type}.png",
                    height=config.plot_scale, width=config.plot_scale * 3,
                    verbose=False, limitsize=False)

        figure.save(f"Figures/{chart_type.title()}s/{folder}/{thisroi}_{chart_type}.svg",
                    height=config.plot_scale, width=config.plot_scale * 3,
                    verbose=False, limitsize=False)

        if config.verbose:
            print(f"Saved {thisroi}_{chart_type}.png")

    @staticmethod
    def load_and_restructure_jsons(config, jsons, combined_df, data_type):
        combined_raw_df = pd.DataFrame()

        # Make a list of significant columns and remove any blank values
        # TODO remove references to histogram to make it more generalisable

        if data_type == 'statistics':
            signif_columns = list(filter(None, [*config.parameter_dict1, "File_name", 'subject']))
        else:
            signif_columns = list(filter(None, [config.histogram_fig_x_facet, config.histogram_fig_y_facet,
                                                "File_name"]))

        for json_file in jsons:
            with open(json_file, 'r') as f:
                current_json = Utils.dict_to_dataframe(json.load(f))

            try:
                current_json = current_json.loc[['Mean', 'Voxels']]
            except KeyError:  # Will raise a KeyError if passing in raw results
                pass

            delimeters = ['_ps', '_raw', '.json']
            json_file_name = os.path.basename(json_file)

            for delimeter in delimeters:
                json_file_name = json_file_name.rsplit(delimeter)[0]

            # Find parameter values for each file_name
            combined_df_search = combined_df.loc[combined_df["File_name"].str.lower() == json_file_name.lower()]
            combined_df_search.columns = combined_df_search.columns.str.lower()

            try:
                for column in signif_columns:
                    current_json[column] = combined_df_search[column.lower()].iloc[0]
            except KeyError:
                pass
            except IndexError:
                continue

            if data_type == 'statistics':
                current_json['subject'] = re.findall('sub-[0-9]*', json_file)[0]

            combined_raw_df = pd.concat([combined_raw_df, current_json])

        # Convert it from wide format into long format
        mean_data = combined_raw_df.iloc[0::2].melt(id_vars=signif_columns, var_name='ROI', value_name='voxel_value')
        voxel_data = combined_raw_df.iloc[1::2].melt(id_vars=signif_columns, var_name='ROI', value_name='voxel_value')

        mean_data['voxel_amount'] = voxel_data['voxel_value']
        mean_data.dropna(inplace=True)  # Drop rows that have NA for voxel value

        return mean_data


class BrainGrid(Figures):
    @classmethod
    def make(cls, combined_results_df):
        indiv_brains_dir = f"{os.getcwd()}/Figures/Brain_images"
        Utils.check_and_make_dir(indiv_brains_dir)
        Utils.check_and_make_dir(f"{os.getcwd()}/Figures/Brain_grids")

        if cls.config.averaging_type == 'Session averaged':
            statistic_labels = cls.config.statistic_options['Session averaged']
            subfolder = 'Session_averaged_results'
        else:
            statistic_labels = cls.config.statistic_options['Participant averaged']
            subfolder = 'Participant_averaged_results'

        for statistic in statistic_labels:
            # If NIFTI files have not been created for statistic, skip creating the figure
            if not glob(f"Overall/NIFTI_ROI/{subfolder}/*{statistic}*"):
                continue

            Utils.check_and_make_dir(f"{os.getcwd()}/Figures/Brain_grids/{statistic}")

            brain_plot_exts = [
                f"_{statistic}.nii.gz",
                f"_{statistic}_within_roi_scaled.nii.gz",
                # f"_{statistic}_mixed_roi_scaled.nii.gz"
            ]

            for base_extension in brain_plot_exts:
                indiv_brain_imgs = cls.setup(combined_results_df, base_extension, statistic, subfolder)

                for img in indiv_brain_imgs:
                    Utils.move_file(img, os.getcwd(), indiv_brains_dir)

            if cls.config.verbose:
                print("\n")

    @classmethod
    def setup(cls, df, base_extension, statistic, subfolder):
        base_ext_clean = os.path.splitext(os.path.splitext(base_extension)[0])[0][1:]
        indiv_brain_imgs = []

        json_array = df['File_name'].unique()

        # Create new list of files which exist in NIFTI_ROI folder
        json_array = [jsn for jsn in json_array if glob(f"Overall/NIFTI_ROI/{subfolder}/{jsn}{base_extension}")]

        critical_params, cell_nums, y_axis_size, x_axis_size = cls.table_setup(df)

        if x_axis_size in (1, 2):
            brain_table_x_size = 32
        else:
            brain_table_x_size = x_axis_size * 16

        brain_table_y_size = 15

        plt.rcParams['figure.figsize'] = brain_table_x_size, brain_table_y_size
        plt.rcParams['figure.subplot.wspace'] = 0.01
        plt.rcParams['figure.subplot.hspace'] = 0.1

        if statistic in ['Mean', 'Median']:
            vmax = vmax_storage = cls.config.brain_fig_value_max
            vmin = cls.config.brain_fig_value_min
        else:
            vmax = vmax_storage = None
            vmin = 0

        if vmax_storage is None:
            vmax_storage = []

        while True:
            if base_extension != f"_{statistic}.nii.gz":
                # For within and mixed scaled tables
                vmax = 100
                vmin = 0
            elif base_extension == f"_{statistic}.nii.gz" and vmax is not None:
                base_ext_clean += "_same_scale"

            if cls.config.verbose:
                print(f"Saving {base_ext_clean} table.")

            indiv_brain_imgs = cls.make_table(base_ext_clean, base_extension, cell_nums, indiv_brain_imgs, json_array,
                                              critical_params, statistic, vmax, vmax_storage, vmin, x_axis_size,
                                              y_axis_size, subfolder)

            if vmax is not None:
                break

            else:
                # Find highest ROI value seen to create figures with the same scale
                vmax_storage = sorted(vmax_storage)[-1]
                vmax = vmax_storage[0]

                if cls.config.verbose:
                    print(
                        f'Maximum ROI value of: {round(vmax_storage[0])} seen for parameter combination: {vmax_storage[1]}. '
                        f'Creating figures with this colourbar limit.')

        return indiv_brain_imgs

    @classmethod
    def make_table(cls, base_ext_clean, base_extension, cell_nums, indiv_brain_imgs, json_array, critical_params,
                   statistic, vmax, vmax_storage, vmin, x_axis_size, y_axis_size, subfolder):
        for file_num, json in enumerate(json_array):
            brain_img, indiv_brain_imgs, dims = cls.save_brain_imgs(json, base_ext_clean, base_extension,
                                                                    vmax, vmax_storage, vmin, indiv_brain_imgs,
                                                                    statistic, subfolder)

            # Import saved image into subplot
            img = mpimg.imread(brain_img)
            plt.subplot(y_axis_size, x_axis_size, cell_nums[file_num] + 1)
            plt.imshow(img)

            ax = plt.gca()
            ax.set_yticks([])  # Remove y-axis ticks
            ax.axes.yaxis.set_ticklabels([])  # Remove y-axis labels

            ax.set_xticks([])  # Remove x-axis ticks
            ax.axes.xaxis.set_ticklabels([])  # Remove x-axis labels

            if critical_params['rows']['order'][file_num] == 0 and len(critical_params['cols']['values']) != 1:
                plt.title(cls.config.brain_table_col_labels + " "
                          + str(critical_params['cols']['values'][critical_params['cols']['order'][file_num]]),
                          fontsize=cls.config.plot_font_size)

            if critical_params['cols']['order'][file_num] == 0 and len(critical_params['rows']['values']) != 1:
                plt.ylabel(cls.config.brain_table_row_labels + " "
                           + str(critical_params['rows']['values'][critical_params['rows']['order'][file_num]]),
                           fontsize=cls.config.plot_font_size)

        cls.label_blank_cell_axes(critical_params, cell_nums, x_axis_size, y_axis_size, dims)

        if cls.config.brain_tight_layout:
            plt.tight_layout()

        plt.savefig(f"Figures/Brain_grids/{statistic}/{base_ext_clean}.png", dpi=cls.config.plot_dpi,
                    bbox_inches='tight')
        plt.savefig(f"Figures/Brain_grids/{statistic}/{base_ext_clean}.svg", dpi=cls.config.plot_dpi,
                    bbox_inches='tight')
        plt.close()

        return indiv_brain_imgs

    @classmethod
    def save_brain_imgs(cls, json, base_ext_clean, base_extension, vmax, vmax_storage, vmin, indiv_brain_imgs,
                        statistic, subfolder):
        # Save brain image using nilearn
        png_path = f"{json}_{base_ext_clean}.png"
        indiv_brain_imgs.append(png_path)

        nifti_path = f"Overall/NIFTI_ROI/{subfolder}/{json}{base_extension}"

        if base_extension == f"_{statistic}.nii.gz" and base_ext_clean.find("_same_scale") == -1:
            # Calculate colour bar limit if not manually set
            brain = nib.load(nifti_path)
            brain = brain.get_fdata()

            vmax = np.nanmax(brain)

            # BE CAREFUL: changes to the vmax_storage list applies outside of the function.
            # If this isn't changed later it is definitely due to efficiency not laziness.
            vmax_storage.append((np.nanmax(brain), json))  # Save vmax to find highest vmax later

        # Correct error caused by nan values by converting them to zeros
        img = nib.load(nifti_path)
        img = nib.Nifti1Image(np.nan_to_num(img.get_fdata()), img.affine, img.header)

        plot = plotting.plot_anat(img,
                                  draw_cross=False, annotate=False, colorbar=True, display_mode='xz',
                                  vmin=vmin, vmax=vmax,
                                  cut_coords=(cls.config.brain_x_coord, cls.config.brain_z_coord),
                                  cmap='inferno')
        plot.savefig(png_path, dpi=cls.config.plot_dpi)
        plot.close()

        with Image.open(png_path) as im:
            width, height = im.size

        return png_path, indiv_brain_imgs, (width, height)

    @classmethod
    def table_setup(cls, df):
        unique_params = []
        cell_nums = []

        for key in cls.config.parameter_dict:
            params = sorted(list(df[key].unique()))  # Sort list of parameters
            params = [str(param) for param in params]  # Convert parameters to strings
            unique_params.append(params)

        plot_values = unique_params  # Get axis values
        axis_titles = list(cls.config.parameter_dict.keys())  # Get axis titles

        critical_params = {'cols': {'param': cls.config.brain_table_cols, 'values': [0], 'order': []},
                           'rows': {'param': cls.config.brain_table_rows, 'values': [0], 'order': []}}

        for axis in critical_params:
            if critical_params[axis]['param'] == '':
                continue
            else:
                critical_params[axis]['values'] = plot_values[
                    axis_titles.index(critical_params[axis]['param'])]  # Sort axis values

        x_axis_size = len(critical_params['cols']['values'])
        y_axis_size = len(critical_params['rows']['values'])

        for file_num, file_name in enumerate(df['File_name'].unique()):
            temp_order_store = []
            file_name_row = df[df['File_name'] == file_name].iloc[0]  # Get the first row of the relevant file name

            for axis in critical_params:
                if critical_params[axis]['values'] != [0]:
                    # Extract parameter from df for file
                    file_param = str(file_name_row[critical_params[axis]['param']])
                    # Work out row/col order
                    critical_params[axis]['order'].append(critical_params[axis]['values'].index(file_param))
                    temp_order_store.append(critical_params[axis]['values'].index(file_param))
                else:
                    critical_params[axis]['order'].append(0)
                    temp_order_store.append(0)

            cell_nums.append(np.ravel_multi_index((temp_order_store[1], temp_order_store[0]),
                                                  (y_axis_size, x_axis_size)))  # Find linear index of figure

        return critical_params, cell_nums, y_axis_size, x_axis_size

    @classmethod
    def label_blank_cell_axes(cls, critical_params, cell_nums, x_axis_size, y_axis_size, dims):
        # Make blank image with the same dimensions as previous images
        img = Image.new("RGB", (dims[0], dims[1]), (255, 255, 255))

        for counter, x_title in enumerate(critical_params['cols']['values']):
            hidden_cell = np.ravel_multi_index((0, counter), (y_axis_size, x_axis_size))

            if hidden_cell not in cell_nums:
                plt.subplot(y_axis_size, x_axis_size, hidden_cell + 1)
                plt.imshow(img)
                cls.make_cell_invisible()

                plt.title(cls.config.brain_table_col_labels + " " + x_title, fontsize=cls.config.plot_font_size)

                if hidden_cell == 0:
                    plt.ylabel(cls.config.brain_table_row_labels + " " + critical_params['rows']['values'][0],
                               fontsize=cls.config.plot_font_size)

        for counter, y_title in enumerate(critical_params['rows']['values']):
            hidden_cell = np.ravel_multi_index((counter, 0), (y_axis_size, x_axis_size))

            if hidden_cell not in cell_nums and hidden_cell != 0:
                plt.subplot(y_axis_size, x_axis_size, hidden_cell + 1)
                plt.imshow(img)
                cls.make_cell_invisible()

                plt.ylabel(cls.config.brain_table_row_labels + " " + y_title, fontsize=cls.config.plot_font_size)

    @staticmethod
    def make_cell_invisible():
        frame = plt.gca()
        frame.axes.get_xaxis().set_ticks([])
        frame.axes.get_yaxis().set_ticks([])
        frame.spines['left'].set_visible(False)
        frame.spines['right'].set_visible(False)
        frame.spines['bottom'].set_visible(False)
        frame.spines['top'].set_visible(False)


class ViolinPlot(Figures):
    @classmethod
    def make(cls, orig_df):
        df = copy.deepcopy(orig_df)

        df = df[(df['index'] != 'Overall') & (df['index'] != 'No ROI')]  # Remove No ROI and Overall rows
        df = df[df['Mean'] > 0]  # Remove any ROIs with value of 0, indicating a failure to fit

        if not cls.config.table_cols == '' and not cls.config.table_rows == '':
            # If parameter to plot for both columns and rows is set then group by both parameters and then sort by mean
            df = df.groupby([cls.config.table_cols, cls.config.table_rows]).apply(lambda x: x.sort_values(['Mean']))

        elif not cls.config.table_cols == '':
            df = df.groupby(cls.config.table_cols).apply(lambda x: x.sort_values(['Mean']))

        elif not cls.config.table_rows == '':
            df = df.groupby(cls.config.table_rows).apply(lambda x: x.sort_values(['Mean']))

        df = df.reset_index(drop=True)  # Reset index to remove grouping

        df['constant'] = 1

        # How much to shift the violin, points and lines
        shift = 0.1

        right_shift = pltn.aes(x=pltn.stage('constant', after_scale='x+shift'))  # shift outward
        left_shift = pltn.aes(x=pltn.stage('constant', after_scale='x-shift'))  # shift inward

        figure = (pltn.ggplot(df)
                  + pltn.aes(x="constant", y="Mean")
                  + pltn.geom_violin(left_shift, na_rm=True, style='left', fill=cls.config.violin_colour, size=0.6)
                  + pltn.geom_boxplot(width=0.1, outlier_alpha=0, fill=cls.config.boxplot_colour, size=0.6)
                  + pltn.xlim(0.4, 1.4)
                  + pltn.ylim(0, None)
                  + pltn.ylab(cls.config.table_x_label)
                  + pltn.xlab("")
                  + pltn.facet_grid(f'{cls.config.table_rows}~{cls.config.table_cols}', drop=True,
                                    labeller="label_both")
                  + pltn.theme_538()  # Set theme
                  + pltn.theme(panel_grid_major_y=pltn.themes.element_line(alpha=1),
                               panel_grid_major_x=pltn.themes.element_line(alpha=0),
                               panel_background=pltn.element_rect(fill="gray", alpha=0.1),
                               axis_text_x=pltn.element_blank(),
                               dpi=cls.config.plot_dpi))

        if cls.config.violin_show_data:
            if cls.config.violin_jitter:
                figure += pltn.geom_jitter(width=0.04, height=0)
            else:
                figure += pltn.geom_point()

        figure.save(f"Figures/Violin_plots/violinplot.png", height=cls.config.plot_scale,
                    width=cls.config.plot_scale * 3,
                    verbose=False, limitsize=False)

        figure.save(f"Figures/Violin_plots/violinplot.svg", height=cls.config.plot_scale,
                    width=cls.config.plot_scale * 3,
                    verbose=False, limitsize=False)

        if cls.config.verbose:
            print(f"Saved violin plot!")


class Barchart(Figures):
    @classmethod
    def setup(cls, df, pool):
        if cls.config.single_roi_fig_x_axis == '':
            raise Exception('Parameter to plot along the x-axes of the barcharts has not been set.')

        Utils.check_and_make_dir("Figures/Barcharts")

        list_rois = list(df['index'].unique())
        chosen_rois = Utils.find_chosen_rois(list_rois, func_name="One region bar chart",
                                             config_region_var=cls.config.regional_fig_rois)
        try:
            chosen_rois
        except NameError:
            if cls.config.verbose:
                print('Skipping barchart creation.')
        else:
            if cls.config.verbose:
                print(f'\n--- Barchart creation ---')

            ylim = 0
            while True:
                iterable = zip(itertools.repeat(cls.make), chosen_rois, itertools.repeat(df),
                               itertools.repeat(ylim), itertools.repeat(cls.figure_save),
                               itertools.repeat(cls.find_axis_limit), itertools.repeat(cls.config))

                if pool:
                    ylim = pool.starmap(Utils.class_method_handler, iterable)

                else:
                    ylim = list(itertools.starmap(Utils.class_method_handler, iterable))

                if any(ylim):
                    ylim.sort(key=lambda x: x[1])

                    if cls.config.verbose:
                        print(f'Maximum y limit of: {round(ylim[-1][1])} seen with ROI: {ylim[-1][2]}. '
                              f'Creating figures with this y limit.\n')

                    ylim = ylim[-1][1]
                else:
                    break

    @staticmethod
    def make(thisroi, df, ylimit, save_function, find_ylim_function, config):
        current_df = df.loc[df['index'] == thisroi]

        current_df = current_df.sort_values([config.single_roi_fig_x_axis])
        current_df = current_df.reset_index(drop=True)  # Reset index to remove grouping
        current_df[config.single_roi_fig_x_axis] = pd.Categorical(current_df[config.single_roi_fig_x_axis],
                                                                  categories=current_df[
                                                                      config.single_roi_fig_x_axis].unique())

        current_df.columns = [c.replace(' ', '_') for c in current_df.columns]
        config.single_roi_fig_x_axis = config.single_roi_fig_x_axis.replace(" ", "_")  # TODO: Comment this
        config.single_roi_fig_colour = config.single_roi_fig_colour.replace(" ", "_")

        figure = (
                pltn.ggplot(current_df)
                + pltn.theme_538()
                + pltn.geom_col(position=pltn.position_dodge(preserve='single', width=0.8), width=0.8, na_rm=True)
                + pltn.geom_errorbar(size=1, position=pltn.position_dodge(preserve='single', width=0.8))
                + pltn.scale_x_discrete(labels=[])
                + pltn.theme(panel_grid_major_x=pltn.element_line(alpha=0),
                             panel_background=pltn.element_rect(fill='white', alpha=.2),
                             axis_title_x=pltn.element_text(weight='bold', color='black', size=20),
                             axis_title_y=pltn.element_text(weight='bold', color='black', size=20),
                             axis_text_y=pltn.element_text(size=20, color='black'),
                             legend_title=pltn.element_text(size=20, weight='bold', color='black', margin={'b': 20}),
                             legend_text=pltn.element_text(size=20, color='black', margin={'l': 5}),
                             legend_entry_spacing=10,
                             legend_key_size=30,
                             subplots_adjust={'right': 0.85},
                             legend_position=(0.9, 0.8),
                             dpi=config.plot_dpi
                             )
                + pltn.geom_text(pltn.aes(y=-.7, label=config.single_roi_fig_x_axis),
                                 color='black', size=20, va='top')
                + pltn.scale_fill_manual(values=config.colorblind_friendly_plot_colours)
        )

        conf_int_string = [x for x in current_df.keys() if 'Conf_Int' in x][0]

        if not config.single_roi_fig_colour == '':
            figure += pltn.aes(x=config.single_roi_fig_x_axis, y='Mean',
                               ymin=f"Mean-{conf_int_string}", ymax=f"Mean+{conf_int_string}",
                               fill=f'factor({config.single_roi_fig_colour})')
        else:
            figure += pltn.aes(x=config.single_roi_fig_x_axis, y='Mean',
                               ymin=f"Mean-{conf_int_string}", ymax=f"Mean+{conf_int_string}")

        figure += pltn.labs(x=config.single_roi_fig_label_x, y=config.single_roi_fig_label_y,
                            fill=config.single_roi_fig_label_fill)

        if ylimit:
            # Set y limit of figure (used to make it the same for every barchart)
            figure += pltn.ylim(None, ylimit)
            thisroi += '_same_ylim'

        returned_ylim = 0
        if ylimit == 0:
            returned_ylim = find_ylim_function(thisroi, figure, 'yaxis')
            folder = 'Different_yaxis'
        elif ylimit != 0:
            folder = 'Same_yaxis'

        save_function(figure, thisroi, folder, 'barchart', config)

        return returned_ylim


class Histogram(Figures):
    @classmethod
    def setup(cls, combined_df, pool):
        Utils.check_and_make_dir("Figures/Histograms")

        list_rois = list(combined_df['index'].unique())
        chosen_rois = Utils.find_chosen_rois(list_rois, func_name="Histogram",
                                             config_region_var=cls.config.regional_fig_rois)

        try:
            chosen_rois
        except NameError:
            if cls.config.verbose:
                print('Skipping histogram creation.')
        else:
            if cls.config.verbose:
                print(f'\n--- Histogram creation ---')

            # Compile a dataframe containing raw values and parameter values for all ROIs and save as combined_raw_df
            jsons = [f"{os.getcwd()}/Overall/Raw_results/{jsn}" for jsn in
                     Utils.find_files("Overall/Raw_results", "json")]
            combined_raw_df = cls.load_and_restructure_jsons(cls.config, jsons, combined_df, data_type='plotting')

            combined_raw_dfs = []
            for roi in chosen_rois:
                combined_raw_dfs.append(
                    cls.get_mean_and_median_for_each_chosen_roi(roi, combined_raw_df, combined_df))

            xlim = 0
            while True:
                iterable = zip(itertools.repeat(cls.make), chosen_rois, combined_raw_dfs,
                               itertools.repeat(xlim), itertools.repeat(cls.figure_save),
                               itertools.repeat(cls.find_axis_limit), itertools.repeat(cls.config))

                if pool:
                    xlim = pool.starmap(Utils.class_method_handler, iterable)
                else:
                    xlim = list(itertools.starmap(Utils.class_method_handler, iterable))

                if any(xlim):
                    try:
                        xlim.remove(None)
                    except ValueError:
                        pass

                    xlim.sort(key=lambda x: x[1])

                    if cls.config.verbose:
                        print(f'Maximum x limit of: {round(xlim[-1][1])} seen with ROI: {xlim[-1][2]}. '
                              f'Creating figures with this x limit.\n')

                    xlim = xlim[-1][1]
                else:
                    break

    @classmethod
    def get_mean_and_median_for_each_chosen_roi(cls, thisroi, combined_raw_df, combined_df):
        # Set up the df for each chosen roi

        if thisroi == "No ROI":
            return pd.DataFrame()
        elif thisroi == "Overall":
            current_df = combined_raw_df.copy()
            current_df['ROI'] = "Overall"
        else:
            current_df = combined_raw_df[combined_raw_df["ROI"] == thisroi].copy()

        current_df = current_df.dropna()  # Drop na values using pandas function, which is faster than plotnines dropna functions

        # Combine both dataframes to find mean and median statistics
        combined_df = combined_df.rename(
            columns={"index": "ROI"})  # Rename column to maintain parity with combined_df column naming convention

        # Make a list of significant columns and remove any blank values
        signif_columns = list(filter(None, [cls.config.histogram_fig_x_facet, cls.config.histogram_fig_y_facet]))
        current_df = current_df.merge(combined_df,
                                      on=['ROI', *signif_columns],
                                      how='left')

        # Keep only the necessary columns
        keys = [*signif_columns, 'ROI', 'voxel_value', 'Mean', 'Median']

        for column in current_df.columns:
            if column not in keys:
                current_df = current_df.drop(columns=column)

        current_df = pd.melt(current_df, id_vars=keys[:-2], var_name="Statistic",
                             value_vars=["Mean", "Median"], value_name="stat_value")  # Put df into tidy format

        if cls.config.histogram_show_mean and not cls.config.histogram_show_median:
            current_df = current_df.loc[current_df["Statistic"] == "Mean"]
        elif cls.config.histogram_show_median and not cls.config.histogram_show_mean:
            current_df = current_df.loc[current_df["Statistic"] == "Median"]

        return current_df

    @staticmethod
    def make(thisroi, combined_raw_df, xlimit, save_function, find_xlim_function, config):
        if combined_raw_df.empty:
            if config.verbose:
                print('INFO: Histograms cannot be made for the No ROI category.')
            return
        else:
            figure = (
                    pltn.ggplot(combined_raw_df, pltn.aes(x="voxel_value"))
                    + pltn.theme_538()
                    + pltn.geom_histogram(binwidth=config.histogram_binwidth, fill=config.histogram_fig_colour,
                                          boundary=0,
                                          na_rm=True)  # Boundary centers the bars, na_rm cancels error from setting an xlimit
                    + pltn.facet_grid(f"{config.histogram_fig_y_facet}~{config.histogram_fig_x_facet}",
                                      drop=True, labeller="label_both")
                    + pltn.labs(x=config.histogram_fig_label_x, y=config.histogram_fig_label_y)
                    + pltn.theme(
                panel_grid_minor_x=pltn.themes.element_line(alpha=0),
                panel_grid_major_x=pltn.themes.element_line(alpha=1),
                panel_grid_major_y=pltn.element_line(alpha=0),
                plot_background=pltn.element_rect(fill="white"),
                panel_background=pltn.element_rect(fill="gray", alpha=0.1),
                axis_title_x=pltn.element_text(weight='bold', color='black', size=20),
                axis_title_y=pltn.element_text(weight='bold', color='black', size=20),
                strip_text_x=pltn.element_text(weight='bold', size=10, color='black'),
                strip_text_y=pltn.element_text(weight='bold', size=10, color='black'),
                axis_text_x=pltn.element_text(size=10, color='black'),
                axis_text_y=pltn.element_text(size=10, color='black'),
                dpi=config.plot_dpi
            )
            )

            # Display mean or median as vertical lines on plot
            if config.histogram_show_mean or config.histogram_show_median:
                figure += pltn.geom_vline(pltn.aes(xintercept="stat_value", color="Statistic"),
                                          size=config.histogram_stat_line_size)
                figure += pltn.scale_color_manual(values=[config.colorblind_friendly_plot_colours[3],
                                                          config.colorblind_friendly_plot_colours[1]])

            # Display legend for mean and median
            if not config.histogram_show_legend:
                figure += pltn.theme(legend_position='none')

            if xlimit:
                # Set y limit of figure (used to make it the same for every barchart)
                figure += pltn.xlim(-1, xlimit)
                thisroi += '_same_xlim'
            else:
                figure += pltn.xlim(-1, None)

            returned_xlim = 0
            if xlimit == 0:
                returned_xlim = find_xlim_function(thisroi, figure, 'xaxis')
                folder = 'Different_xaxis'
            elif xlimit != 0:
                folder = 'Same_xaxis'

            # Suppress Pandas warning about alignment of non-concatenation axis
            warnings.simplefilter(action='ignore', category=FutureWarning)

            save_function(figure, thisroi, folder, 'histogram', config)

            warnings.simplefilter(action='default', category=FutureWarning)

            return returned_xlim


class CompareOutputs(Figures):
    current_df = None  # TODO: See what this class could be used for

    @classmethod
    def run(cls, config):
        df, labels = cls.setup_df(config)
        cls.Make_scatter(df, labels, config)

    @classmethod
    def setup_df(cls, config):
        dfs = {}
        labels = []
        for x in range(2):
            directory = Utils.file_browser(title='Select the directory output by the fRAT')

            with open(f"{directory}/Summarised_results/combined_results.json", "r") as results:
                data = json.load(results)
                dfs[x] = pd.DataFrame(data)
                labels.append(os.path.basename(directory))

                rois = sorted({d['index'] for d in data})  # Using set returns only unique values
            # TODO: run a check to make sure same ROIs, check they have same critical parameters and same parameter space
        dfm = dfs[0].merge(dfs[1], how='outer',
                           on=['index', config.histogram_fig_y_facet, config.histogram_fig_x_facet])

        return dfm, labels

    @classmethod
    def Make_scatter(cls, df, labels, config):
        figure = (
                pltn.ggplot(df, pltn.aes(x="Mean_x", y="Mean_y"))
                + pltn.theme_538()
                + pltn.geom_point()
                + pltn.facet_grid(f"{config.histogram_fig_y_facet}~{config.histogram_fig_x_facet}",
                                  drop=True, labeller="label_both")
                + pltn.geom_smooth(method='lm')
                + pltn.labs(x=labels[1], y=labels[0])  # TODO Check these labels are the right way round
                + pltn.theme(
            panel_grid_minor_x=pltn.themes.element_line(alpha=0),
            panel_grid_major_x=pltn.themes.element_line(alpha=1),
            panel_grid_major_y=pltn.element_line(alpha=0),
            plot_background=pltn.element_rect(fill="white"),
            panel_background=pltn.element_rect(fill="gray", alpha=0.1),
            axis_title_x=pltn.element_text(weight='bold', color='black', size=20),
            axis_title_y=pltn.element_text(weight='bold', color='black', size=20),
            strip_text_x=pltn.element_text(weight='bold', size=10, color='black'),
            strip_text_y=pltn.element_text(weight='bold', size=10, color='black'),
            axis_text_x=pltn.element_text(size=10, color='black'),
            axis_text_y=pltn.element_text(size=10, color='black'),
            dpi=config.plot_dpi
        )
        )

        figure.save(f"TEST.png", height=config.plot_scale,
                    width=config.plot_scale * 3,
                    verbose=False, limitsize=False)
